"""Implement unit test for the Complex class."""

from math import isclose, sqrt
from hypothesis import given
from hypothesis.strategies import integers, floats, tuples

from complex_class import Complex

VALUE_MIN, VALUE_MAX = -100, 100
cint_tuple1 = tuples(
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
)
cint_tuple2 = tuples(
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
)


@given(reim=cint_tuple1)
def test_init(reim):
    re, im = reim
    z = Complex(re, im)
    assert isclose(float(re), z.real)
    assert isclose(float(im), z.imag)


@given(reim=cint_tuple1)
def test_absvalue(reim):
    re, im = reim
    z = Complex(re, im)
    ref = sqrt(re ** 2 + im ** 2)
    assert isclose(ref, z.abs())


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_add(reim1, reim2):
    # setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 + re2
    ref_im = im1 + im2

    # result
    z_res = z1 + z2

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(x_int=integers(min_value=VALUE_MIN, max_value=VALUE_MAX), reim=cint_tuple1)
def test_radd(x_int, reim):
    # Setup
    re, im = reim
    z = Complex(re, im)

    # reference
    ref_re = x_int + re
    ref_im = im

    # result
    z_res = x_int + z

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_iadd(reim1, reim2):
    # setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 + re2
    ref_im = im1 + im2

    # result
    z1 += z2

    # verify
    assert isclose(float(ref_re), z1.real)
    assert isclose(float(ref_im), z1.imag)


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_sub(reim1, reim2):
    # Setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 - re2
    ref_im = im1 - im2

    # result
    z_res = z1 - z2

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(x_int=integers(min_value=VALUE_MIN, max_value=VALUE_MAX), reim=cint_tuple1)
def test_rsub(x_int, reim):
    # Setup
    re, im = reim
    z = Complex(re, im)

    # reference
    ref_re = x_int - re
    ref_im = -im

    # result
    z_res = x_int - z

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_isub(reim1, reim2):
    # Setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 - re2
    ref_im = im1 - im2

    # result
    z1 -= z2

    # verify
    assert isclose(float(ref_re), z1.real)
    assert isclose(float(ref_im), z1.imag)


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_mul(reim1, reim2):
    # Setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 * re2 - im1 * im2
    ref_im = re1 * im2 + im1 * re2

    # result
    z_res = z1 * z2

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(x_int=integers(min_value=VALUE_MIN, max_value=VALUE_MAX), reim=cint_tuple1)
def test_rmul(x_int, reim):
    # Setup
    re, im = reim
    z = Complex(re, im)

    # reference
    ref_re = x_int * re
    ref_im = x_int * im

    # result
    z_res = x_int * z

    # verify
    assert isclose(float(ref_re), z_res.real)
    assert isclose(float(ref_im), z_res.imag)


@given(reim1=cint_tuple1, reim2=cint_tuple2)
def test_imul(reim1, reim2):
    # Setup
    re1, im1 = reim1
    z1 = Complex(re1, im1)
    re2, im2 = reim2
    z2 = Complex(re2, im2)

    # reference
    ref_re = re1 * re2 - im1 * im2
    ref_im = re1 * im2 + im1 * re2

    # result
    z1 *= z2

    # verify
    assert isclose(float(ref_re), z1.real)
    assert isclose(float(ref_im), z1.imag)
