"""Implement a class for complex numbers."""


class Complex:
    """Custom complex number class."""

    def __init__(self, real=0, imag=0):
        self.real, self.imag = self.__same_numeric_type(re=real, im=imag)

    @classmethod
    def make_Complex(cls, other):
        if (type(other) is int) or (type(other) is float):
            return cls(real=other, imag=0)
        elif type(other) is Complex:
            return other
        else:
            raise ValueError("Invalid type passed to Complex constructor.")

    def __make_float(self, re, im):
        if type(re) != type(im):
            if any((type(re) == float, type(im) == float)):
                myre, myim = float(re), float(im)
        return myre, myim

    def __same_numeric_type(self, re, im):
        """Assert that real and imag part are of the same numerical type.

        If different, meaning any of (real, imag) is of type float while
        the other is not convert both to floats.
        """
        try:
            re + 1, im + 1
        except TypeError:
            print(
                """Input values of real or imag part are
                   not of type 'int' or 'float'."""
            )
            raise
        return self.__make_float(im, re)

    def __repr__(self):
        """Printing a complex number."""
        import math

        copysign = math.copysign(1, self.imag)
        sign = "-" if str(copysign)[0] == "-" else "+"
        return "(%s%s%sj)" % (self.real, sign, int(copysign) * self.imag)

    def __add__(self, other):
        """Adding a (complex) number to the current object.

        This magic is used for a call of the form:

        complex_object + number_object

        where number_object can be `int` or `float`.
        """
        myother = Complex.make_Complex(other)
        re = self.real + myother.real
        im = self.imag + myother.imag
        return Complex(real=re, imag=im)

    def __radd__(self, other):
        """Reverse addition order.

        This magic calls __add__ but makes it possible to
        also invoke calls of the form

        number_object  + complex_object

        where number_object can be `int` or `float`.
        """
        return self.__add__(other)

    def __iadd__(self, other):
        """Adding a (complex) number to the current object in-place."""
        return self.__add__(other)

    def __sub__(self, other):
        """Subtracting a (complex) number from the current object."""
        myother = Complex.make_Complex(other)
        re = self.real - myother.real
        im = self.imag - myother.imag
        return Complex(real=re, imag=im)

    def __rsub__(self, other):
        """Reverse subtraction order.

        See docstring of __radd__ magic.

        Careful:
        -------
        Contrary to __radd__ and __rmul__ the order now actually
        is of importance.
        """
        myother = Complex.make_Complex(other)
        return myother.__sub__(self)

    def __isub__(self, other):
        """Subtract a (complex) number from the current object in-place."""
        return self.__sub__(other)

    def __mul__(self, other):
        """Multiplying a (complex) number with the current object."""
        myother = Complex.make_Complex(other)
        re = self.real * myother.real - self.imag * myother.imag
        im = self.real * myother.imag + self.imag * myother.real
        return Complex(real=re, imag=im)

    def __rmul__(self, other):
        """Reverse multiplication order.

        See docstring of __radd__ magic.
        """
        return self.__mul__(other)

    def __imul__(self, other):
        """Multiply a (complex) number with the current object in-place."""
        return self.__mul__(other)

    def __truediv__(self, other):
        """Dividing current object by a (complex) number"""
        myother = Complex.make_Complex(other)
        denom = myother.real ** 2 + myother.imag ** 2
        re = (self.real * myother.real + self.imag * myother.imag) / denom
        im = (self.imag * myother.real - self.real * myother.imag) / denom
        return Complex(real=re, imag=im)

    def __rtruediv__(self, other):
        """Inverse division order.

        Careful:
        -------
        Contrary to __radd__ and __rmul__ the order now actually
        is of importance.
        """
        myother = Complex.make_Complex(other)
        return myother.__truediv__(self)

    def __itruediv__(self, other):
        """Dividing current object by a (complex) number in-place."""
        myother = Complex.make_Complex(other)
        return self.__truediv__(myother)

    def __pos__(self):
        """Add + sign to complex number object."""
        return Complex(self.real, self.imag)

    def __neg__(self):
        """Add - sign to complex number object."""
        return Complex(-self.real, -self.imag)

    def conjg(self):
        """Complex conjugation of a complex number."""
        return Complex(real=self.real, imag=-self.imag)

    def abs(self):
        """Absolute value of a complex number."""
        return (self.real ** 2 + self.imag ** 2) ** 0.5
