"""Implement unit test for the Complex class."""

from math import isclose, sqrt
from hypothesis import given
from hypothesis.strategies import integers, floats, tuples

from complex_class import Complex

VALUE_MIN, VALUE_MAX = -100, 100
cint_tuple1 = tuples(
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
).filter(lambda x: x[0] != 0 or x[1] != 0)
cint_tuple2 = tuples(
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
).filter(lambda x: x[0] != 0 or x[1] != 0)

cfloat_tuple1 = tuples(
    floats(min_value=VALUE_MIN, max_value=VALUE_MAX),
    floats(min_value=VALUE_MIN, max_value=VALUE_MAX),
).filter(lambda x: abs(x[0]) > 0 or abs(x[1]) > 0)
cfloat_tuple2 = tuples(
    floats(min_value=VALUE_MIN, max_value=VALUE_MAX),
    floats(min_value=VALUE_MIN, max_value=VALUE_MAX),
).filter(lambda x: abs(x[0]) > 0 or abs(x[1]) > 0)

cintfloat_tuple = tuples(
    integers(min_value=VALUE_MIN, max_value=VALUE_MAX),
    floats(min_value=VALUE_MIN, max_value=VALUE_MAX),
).filter(lambda x: x[0] != 0 or abs(x[1]) > 0)


class ComplexArithmetic(object):
    """Wrapper class for the Complex class used for testing arithmetic ops."""

    def __init__(self, complex_obj, numerical_obj, arithmetic_op):
        self.complex_obj = complex_obj
        self.numerical_obj = numerical_obj
        self.arithmetic_op = arithmetic_op
        self.__results = {
            "add": lambda x: self.complex_obj.__add__(x),
            "iadd": lambda x: self.complex_obj.__iadd__(x),
            "radd": lambda x: self.complex_obj.__radd__(x),
            "sub": lambda x: self.complex_obj.__sub__(x),
            "isub": lambda x: self.complex_obj.__isub__(x),
            "rsub": lambda x: self.complex_obj.__rsub__(x),
            "mul": lambda x: self.complex_obj.__mul__(x),
            "imul": lambda x: self.complex_obj.__imul__(x),
            "rmul": lambda x: self.complex_obj.__rmul__(x),
            "truediv": lambda x: self.complex_obj.__truediv__(x),
            "itruediv": lambda x: self.complex_obj.__itruediv__(x),
            "rtruediv": lambda x: self.complex_obj.__rtruediv__(x),
        }

    def get_result(self):
        """Get the result of a arithmetic operation."""
        return self.__results[self.arithmetic_op](self.numerical_obj)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        del self.__results


def evaluate_arithmetic(z1, z2):
    """Test values returned arithmetic operations."""
    # addition
    for op in ("add", "iadd"):
        with ComplexArithmetic(z1, z2, op) as add:
            result = add.get_result()
            assert isclose(result.real, z1.real + z2.real)
            assert isclose(result.imag, z1.imag + z2.imag)

    # subtraction
    for op in ("sub", "isub"):
        with ComplexArithmetic(z1, z2, op) as sub:
            result = sub.get_result()
            assert isclose(result.real, z1.real - z2.real)
            assert isclose(result.imag, z1.imag - z2.imag)

    # multiplication
    for op in ("mul", "imul"):
        with ComplexArithmetic(z1, z2, op) as mul:
            result = mul.get_result()
            assert isclose(result.real, z1.real * z2.real - z1.imag * z2.imag)
            assert isclose(result.imag, z1.real * z2.imag + z1.imag * z2.real)

    # division
    for op in ("truediv", "itruediv"):
        with ComplexArithmetic(z1, z2, op) as div:
            z = div.get_result()
            denom = z2.real ** 2 + z2.imag ** 2
            assert isclose(z.real, (z1.real * z2.real + z1.imag * z2.imag) / denom)
            assert isclose(z.imag, (z1.imag * z2.real - z1.real * z2.imag) / denom)


def evaluate_reversed_arithmetic(x, z):
    """Test values returned by reversed arithmetic operations."""
    # addition
    with ComplexArithmetic(complex_obj=z, numerical_obj=x, arithmetic_op="radd") as add:
        result = add.get_result()
        assert isclose(result.real, x + z.real)
        assert isclose(result.imag, z.imag)

    # subtraction
    with ComplexArithmetic(complex_obj=z, numerical_obj=x, arithmetic_op="rsub") as add:
        result = add.get_result()
        assert isclose(result.real, x - z.real)
        assert isclose(result.imag, -z.imag)

    # multiplication
    with ComplexArithmetic(complex_obj=z, numerical_obj=x, arithmetic_op="rmul") as add:
        result = add.get_result()
        assert isclose(result.real, x * z.real)
        assert isclose(result.imag, x * z.imag)

    # division
    with ComplexArithmetic(
        complex_obj=z, numerical_obj=x, arithmetic_op="rtruediv"
    ) as add:
        result = add.get_result()
        denominator = z.real ** 2 + z.imag ** 2
        assert isclose(result.real, x * z.real / denominator)
        assert isclose(result.imag, -x * z.imag / denominator)


def types_arithmetic(z1, z2):
    """Test types returned by arithmetic operations."""
    list_of_operations = (
        "add",
        "iadd",
        "radd",
        "sub",
        "isub",
        "rsub",
        "mul",
        "imul",
        "rmul",
        "truediv",
        "itruediv",
        "rtruediv",
    )
    for op in list_of_operations:
        with ComplexArithmetic(z1, z2, op) as result_of_op:
            result = result_of_op.get_result()
            assert type(result) is Complex


# testing initialisation
def test_init_default():
    """Test initialisation with default args."""
    c = Complex()
    assert c.real == 0
    assert c.imag == 0
    assert type(c) is Complex


@given(
    realimag_int=cint_tuple1,
    realimag_float=cfloat_tuple1,
    realimag_intfloat=cintfloat_tuple,
)
def test_init_type(realimag_int, realimag_float, realimag_intfloat):
    """Test if in case of mixed args (int, float) real and imag part are both float"""
    c_intint = Complex(*realimag_int)
    c_intfloat = Complex(*realimag_intfloat)
    c_floatint = Complex(*realimag_intfloat[::-1])
    c_floatfloat = Complex(*realimag_float)

    # return type of object itself
    assert type(c_intint) is Complex
    assert type(c_intfloat) is Complex
    assert type(c_floatint) is Complex
    assert type(c_floatfloat) is Complex

    # types of real and imag part
    assert type(c_intint.real) is int
    assert type(c_intint.imag) is int
    assert type(c_intfloat.real) is float
    assert type(c_intfloat.imag) is float
    assert type(c_floatint.real) is float
    assert type(c_floatint.imag) is float
    assert type(c_floatfloat.real) is float
    assert type(c_floatfloat.imag) is float


@given(x=integers(min_value=VALUE_MIN, max_value=VALUE_MAX))
def test_init_one_int_arg(x):
    """Test initialisation with only one integer arg."""
    c_real_default = Complex(x)
    c_real = Complex(x, 0)
    c_imag = Complex(0, x)

    # type of objects
    assert type(c_real_default) is Complex
    assert type(c_real) is Complex
    assert type(c_imag) is Complex

    assert c_real.real == x
    assert c_real.imag == 0

    assert c_real_default.real == c_real.real
    assert c_real_default.imag == c_real.imag

    assert c_imag.real == 0
    assert c_imag.imag == x


@given(x=floats(min_value=VALUE_MIN, max_value=VALUE_MAX))
def test_init_one_float_arg(x):
    """Test initialisation with only one float arg."""
    c_real_default = Complex(x)
    c_real = Complex(x, 0.0)
    c_imag = Complex(0.0, x)

    # types of objects
    assert type(c_real_default) is Complex
    assert type(c_real) is Complex
    assert type(c_imag) is Complex

    assert isclose(c_real.real, x)
    assert isclose(c_real.imag, 0.0)

    assert isclose(c_real_default.real - c_real.real, 0.0)
    assert isclose(c_real_default.imag - c_real.imag, 0.0)

    assert isclose(c_imag.real, 0.0)
    assert isclose(c_imag.imag, x)


@given(realimag=cint_tuple1)
def test_init_two_int_args(realimag):
    """Test initialisation with two integer args."""
    c = Complex(*realimag)
    assert (c.real == realimag[0]) and (c.imag == realimag[1])


@given(realimag=cfloat_tuple1)
def test_init_two_float_args(realimag):
    """Test initialisation with two integer args."""
    c = Complex(*realimag)
    assert isclose(c.real - realimag[0], 0.0)
    assert isclose(c.imag - realimag[1], 0.0)


@given(realimag=cfloat_tuple1)
def test_sign_change(realimag):
    """Test initialisation with two integer args."""
    c = Complex(*realimag)
    c_pos = +c
    c_neg = -c
    assert isclose(c.real - c_pos.real, 0.0)
    assert isclose(c.imag - c_pos.imag, 0.0)
    assert isclose(c.real + c_neg.real, 0.0)
    assert isclose(c.imag + c_neg.imag, 0.0)


# testing arithmetic operations
@given(realimag=cint_tuple1)
def test_integer_uniform_arithmetic_types(realimag):
    """Test return types of arithmetic for !!!two!!! float-based args."""
    types_arithmetic(z1=Complex(*realimag), z2=Complex(*realimag))


@given(realimag=cfloat_tuple1)
def test_float_uniform_arithmetic_types(realimag):
    """Test return types of arithmetic for !!!two!!! float-based args."""
    types_arithmetic(z1=Complex(*realimag), z2=Complex(*realimag))


@given(realimag1=cint_tuple1, realimag2=cint_tuple2)
def test_integer_uniform_arithmetic_values(realimag1, realimag2):
    """Test return values of arithmetic for !!!two!!! integer-based args."""
    evaluate_arithmetic(z1=Complex(*realimag1), z2=Complex(*realimag2))


@given(realimag1=cfloat_tuple1, realimag2=cfloat_tuple2)
def test_float_uniform_arithmetic_values(realimag1, realimag2):
    """Test return values of arithmetic for !!!two!!! float-based args."""
    evaluate_arithmetic(z1=Complex(*realimag1), z2=Complex(*realimag2))


@given(
    x=integers(min_value=VALUE_MIN, max_value=VALUE_MAX), realimag_float=cfloat_tuple2
)
def test_mixed_with_int_arithmetic_values(x, realimag_float):
    evaluate_reversed_arithmetic(x, Complex(*realimag_float))


@given(x=floats(min_value=VALUE_MIN, max_value=VALUE_MAX), realimag_float=cfloat_tuple2)
def test_mixed_with_float_arithmetic_values(x, realimag_float):
    evaluate_reversed_arithmetic(x, Complex(*realimag_float))


@given(realimag=cfloat_tuple1)
def test_complex_conjugation(realimag):
    """Test complex conjugation."""
    c = Complex(*realimag)
    c_conj = c.conjg()
    assert type(c_conj) is Complex
    assert isclose(c.real - c_conj.real, 0.0)
    assert isclose(c.imag + c_conj.imag, 0.0)


@given(realimag=cfloat_tuple1)
def test_absolute_value(realimag):
    """Test absolute values."""
    c = Complex(*realimag)
    abs_value = c.absval()
    assert type(abs_value) is float
    assert isclose(abs_value, sqrt(c.real ** 2 + c.imag ** 2))
